#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from keras.callbacks import BaseLogger, CallbackList, History, ProgbarLogger
import numpy as np

from .datasets import InMemoryDataset, GeneratorDataset
from .model_wrappers import OracleWrapper
from .samplers import AdaptiveAdditiveSmoothingSampler, \
    AdditiveSmoothingSampler, ModelSampler, LSTMSampler
from .reweighting import BiasedReweightingPolicy
from .utils.functional import ___, compose, partial


class _BaseImportanceTraining(object):
    def __init__(self, model):
        """Abstract base class for training a model using importance sampling.
        
        Arguments
        ---------
            model: The Keras model to train
        """
        # Wrap and transform the model so that it outputs the importance scores
        # and can be used in an importance sampling training scheme
        self.original_model = model
        self.model = OracleWrapper(model, self.reweighting)

    @property
    def reweighting(self):
        """The reweighting policy that controls the bias of the estimation when
        using importance sampling."""
        raise NotImplementedError()

    def sampler(self, dataset):
        """Create a new sampler to sample from the given dataset using
        importance sampling."""
        raise NotImplementedError()

    def fit(self, x, y, batch_size=32, epochs=1, verbose=1, callbacks=None,
            validation_split=0.0, validation_data=None, steps_per_epoch=None):
        """Create an `InMemoryDataset` instance with the given data and train
        the model using importance sampling for a given number of epochs.

        Arguments
        ---------
            x: Numpy array of training data
            y: Numpy array of target data
            batch_size: int, number of samples per gradient update
            epochs: int, number of times to iterate over the entire
                    training set
            verbose: {0, >0}, whether to employ the progress bar Keras
                     callback or not
            callbacks: list of Keras callbacks to be called during training
            validation_split: float in [0, 1), percentage of data to use for
                              evaluation
            validation_data: tuple of numpy arrays, Data to evaluate the
                             trained model on without ever training on them
            steps_per_epoch: int or None, number of gradient updates before
                             considering an epoch has passed
        Returns
        -------
            A Keras `History` object that contains information collected during
            training.
        """
        # Create two data tuples from the given x, y, validation_*
        if validation_data is not None:
            x_train, y_train = x, y
            x_test, y_test = validation_data

        elif validation_split > 0:
            assert validation_split < 1, "100% of the data used for testing"
            n = int(round(validation_split * len(x)))
            idxs = np.arange(len(x))
            np.random.shuffle(idxs)
            x_train, y_train = x[idxs[n:]], y[idxs[n:]]
            x_test, y_test = x[idxs[:n]], y[idxs[:n]]

        else:
            x_train, y_train = x, y
            x_test, y_test = np.empty(shape=(0, 1)), np.empty(shape=(0, 1))

        # Make the dataset to train on
        dataset = InMemoryDataset(
            x_train,
            y_train,
            x_test,
            y_test,
            categorical=False  # this means use the targets as is
        )

        return self.fit_dataset(
            dataset=dataset,
            batch_size=batch_size,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            verbose=verbose,
            callbacks=callbacks
        )

    def fit_generator(self, train, steps_per_epoch, batch_size=32,
                      epochs=1, verbose=1, callbacks=None,
                      validation_data=None, validation_steps=None):
        """Create a GeneratorDataset instance and train the model using
        importance sampling for a given number of epochs.

        NOTICE: This method may not be supported by all importance training
        classes and may result in NotImplementedError()

        Arguments
        ---------
            train: A generator that returns tuples (inputs, targets)
            steps_per_epoch: int, number of gradient updates before considering
                             an epoch has passed
            batch_size: int, the number of samples per gradient update (ideally
                             set to the number of items returned by the
                             generator at each call)
            epochs: int, multiplied by steps_per_epoch denotes the total number
                    of gradient updates
            verbose: {0, >0}, whether to use the progress bar Keras callback
            validation_data: generator or tuple (inputs, targets)
            validation_steps: None or int, used only if validation_data is a
                              generator
        """
        # Create the validation data to pass to the GeneratorDataset
        if validation_data is not None:
            if isinstance(validation_data, (tuple, list)):
                test = validation_data
                test_len = None
            else:
                test = validation_data
                test_len = validation_steps * batch_size
        else:
            test = (np.empty(shape=(0, 1)), np.empty(shape=(0, 1)))
            test_len = None

        dataset = GeneratorDataset(train, test, test_len)

        return self.fit_dataset(
            dataset=dataset,
            batch_size=batch_size,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            verbose=verbose,
            callbacks=callbacks
        )

    def fit_dataset(self, dataset, steps_per_epoch=None, batch_size=32,
                    epochs=1, verbose=1, callbacks=None):
        """Train the model on the given dataset for a given number of epochs.

        Arguments
        ---------
            dataset: Instance of `BaseDataset` that provides the data
                     to train on.
            steps_per_epoch: int or None, number of gradient updates before
                             considering an epoch has passed. If None it is set
                             to be `len(dataset.train_data) / batch_size`.
            batch_size: int, number of samples per gradient update
            epochs: int, number of times to iterate `steps_per_epoch` times
            verbose: {0, >0}, whether to employ the progress bar Keras
                     callback or not
            callbacks: list of Keras callbacks to be called during training
        """
        # Set steps_per_epoch properly
        if steps_per_epoch is None:
            steps_per_epoch = len(dataset.train_data) / batch_size

        # Create the callbacks list
        self.history = History()
        callbacks = [BaseLogger()] + (callbacks or []) + [self.history]
        if verbose > 0:
            callbacks += [ProgbarLogger(count_mode="steps")]
        callbacks = CallbackList(callbacks)
        #TODO: Should we be making it possible to call back a different model
        #      than self.model.model?
        callbacks.set_model(self.model.model)
        callbacks.set_params({
            "epochs": epochs,
            "steps": steps_per_epoch,
            "verbose": verbose,
            "do_validation": len(dataset.test_data) > 0,
            "metrics": self._get_metric_names() + [
                "val_" + n for n in self._get_metric_names()
            ]
        })

        # Create the sampler
        sampler = self.sampler(dataset)

        # Start the training loop
        epoch = 0
        self.model.model.stop_training = False
        callbacks.on_train_begin()
        while epoch < epochs:
            callbacks.on_epoch_begin(epoch)
            for step in range(steps_per_epoch):
                batch_logs = {"batch": step, "size": batch_size}
                callbacks.on_batch_begin(step, batch_logs)

                # Importance sampling is done here
                idxs, (x, y), w = sampler.sample(batch_size)
                # Train on the sampled data
                loss, metrics, scores = self.model.train_batch(x, y, w)
                # Update the sampler
                sampler.update(idxs, scores)

                values = map(lambda x: x.mean(), [loss] + metrics)
                for l, o in zip(self._get_metric_names(), values):
                    batch_logs[l] = o
                callbacks.on_batch_end(step, batch_logs)

            # Evaluate now that an epoch passed
            epoch_logs = {}
            if len(dataset.test_data) > 0:
                val = self.model.evaluate(
                    *dataset.test_data[:],
                    batch_size=batch_size
                )
                epoch_logs = {
                    "val_" + l: o
                    for l, o in zip(self._get_metric_names(), val)
                }
            callbacks.on_epoch_end(epoch, epoch_logs)
            if self.model.model.stop_training:
                break
            epoch += 1
        callbacks.on_train_end()

        return self.history

    def _get_metric_names(self):
        metrics = self.original_model.metrics or []
        return (
            ["loss"] +
            map(str, metrics) +
            ["score"]
        )


class ImportanceTraining(_BaseImportanceTraining):
    """Train a model with exact importance sampling using the loss as
    importance.
    
    Arguments
    ---------
        model: The Keras model to train
        k: float in (-oo, 1], controls the bias of the training that focuses
           the network on the hard examples (see paper)
        smooth: float, influences the sampling distribution towards uniform by
                adding `smooth` to all probabilities and renormalizing
        adaptive_smoothing: bool, If True the `smooth` argument is a percentage
                            of the average training loss
        presample: int, the number of samples to presample for scoring
        forward_batch_size: int, the batch size to use for the forward pass
                            during scoring
    """
    def __init__(self, model, k=0.5, smooth=0.0, adaptive_smoothing=False,
                 presample=1024, forward_batch_size=128):
        # Create the reweighting policy
        self._reweighting = BiasedReweightingPolicy(k)

        # Call the parent to wrap the model
        super(ImportanceTraining, self).__init__(model)

        # Create the sampler factory, the workhorse of the whole deal :-)
        adaptive_smoothing_factory = partial(
            AdaptiveAdditiveSmoothingSampler, ___, smooth
        )
        additive_smoothing_factory = partial(
            AdditiveSmoothingSampler, ___, smooth
        )
        self._sampler = partial(
            ModelSampler,
            ___,
            self.reweighting,
            self.model,
            large_batch=presample,
            forward_batch_size=forward_batch_size
        )
        if adaptive_smoothing and smooth > 0:
            self._sampler = compose(adaptive_smoothing_factory, self._sampler)
        elif smooth > 0:
            self._sampler = compose(additive_smoothing_factory, self._sampler)

    @property
    def reweighting(self):
        return self._reweighting

    def sampler(self, dataset):
        return self._sampler(dataset)


class ApproximateImportanceTraining(_BaseImportanceTraining):
    """Train a model with importance sampling using an LSTM with a class
    embedding to predict the importance of the training samples.
    
    Arguments
    ---------
        model: The Keras model to train
        k: float in (-oo, 1], controls the bias of the training that focuses
           the network on the hard examples (see paper)
        smooth: float, influences the sampling distribution towards uniform by
                adding `smooth` to all probabilities and renormalizing
        adaptive_smoothing: bool, If True the `smooth` argument is a percentage
                            of the average training loss
        presample: int, the number of samples to presample for scoring
    """
    def __init__(self, model, k=0.5, smooth=0.0, adaptive_smoothing=False,
                 presample=2048):
        # Create the reweighting policy
        self._reweighting = BiasedReweightingPolicy(k)

        # Call the parent to wrap the model
        super(ApproximateImportanceTraining, self).__init__(model)

        # Create the sampler factory, the workhorse of the whole deal :-)
        adaptive_smoothing_factory = partial(
            AdaptiveAdditiveSmoothingSampler, ___, smooth
        )
        additive_smoothing_factory = partial(
            AdditiveSmoothingSampler, ___, smooth
        )
        self._sampler = partial(
            LSTMSampler,
            ___,
            self.reweighting,
            presample=presample
        )
        if adaptive_smoothing and smooth > 0:
            self._sampler = compose(adaptive_smoothing_factory, self._sampler)
        elif smooth > 0:
            self._sampler = compose(additive_smoothing_factory, self._sampler)

    @property
    def reweighting(self):
        return self._reweighting

    def sampler(self, dataset):
        return self._sampler(dataset)

    def fit_generator(*args, **kwargs):
        raise NotImplementedError("ApproximateImportanceTraining doesn't "
                                  "support generator training")



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Accelerate training of neural networks using importance sampling."""

__author__ = "Angelos Katharopoulos"
__copyright__ = "Copyright (c) 2017 Idiap Research Institute"
__license__ = "MIT"
__maintainer__ = "Angelos Katharopoulos"
__email__ = "angelos.katharopoulos@idiap.ch"
__url__ = "http://www.idiap.ch/~katharas/importance-sampling/"
__version__ = "0.2"

from training import ImportanceTraining, ApproximateImportanceTraining



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from functools import partial

from blinker import signal
from keras.layers import Dense, Embedding, Flatten, Input, LSTM, Masking, \
    concatenate
from keras.models import Model
import numpy as np


def _get_dataset_length(dset, default=1):
    """Return the dataset's training data length and in case the dataset is
    uncountable return a defalt value."""
    try:
        return len(dset.train_data)
    except RuntimeError:
        return default


class BaseSampler(object):
    """BaseSampler denotes the interface for all the samplers.

    Samplers should provide the rest of the program with data points to train
    on and corresponding relative weights."""
    def __init__(self, dataset, reweighting):
        self.dataset = dataset
        self.reweighting = reweighting

    def _get_samples_with_scores(self, batch_size):
        """Child classes should implement this method.

        Arguments
        ---------
        batch_size: int
                    Return at least that many samples
        
        Return
        ------
        idxs: array
              The indices of some samples in the dataset
        scores: array or None
                The predicted importance scores for the corresponding idxs or
                None for uniform sampling
        xy: tuple or None
            Optionally return the data for the corresponding idxs
        """
        raise NotImplementedError()


    def sample(self, batch_size):
        # Get the importance scores of some samples
        idxs1, scores, xy = self._get_samples_with_scores(batch_size)

        # Sample from the available ones
        p = scores / scores.sum() if scores is not None else None
        idxs2 = np.random.choice(len(idxs1), batch_size, p=p)
        w = self.reweighting.sample_weights(idxs2, scores)

        # Make sure we have the data
        if xy is None:
            xy = self.dataset.train_data[idxs1[idxs2]]
        else:
            x, y = xy
            xy = (x[idxs2], y[idxs2])

        scores = scores[idxs2] if scores is not None else np.ones(batch_size)
        signal("is.sample").send({
            "idxs": idxs1,
            "xy": xy,
            "w": w,
            "predicted_scores": scores
        })
        return idxs1[idxs2], xy, w

    def update(self, idxs, results):
        pass


class UniformSampler(BaseSampler):
    """UniformSampler is the simplest possible sampler which samples the
    dataset uniformly."""
    def __init__(self, dataset, reweighting):
        super(UniformSampler, self).__init__(dataset, reweighting)
        # Basically if we don't know the length the indices don't matter so
        # sample batch_size 0s.
        self.idxs = np.arange(_get_dataset_length(self.dataset, default=1))

    def _get_samples_with_scores(self, batch_size):
        return (
            self.idxs,
            None,
            None
        )


class ModelSampler(BaseSampler):
    """ModelSampler uses a model to score the samples and then performs
    importance sampling based on those scores.

    It can be used to implement several training pipelines where the scoring
    model is separately trained or is sampled from the main model or is the
    main model."""
    def __init__(self, dataset, reweighting, model, large_batch=1024,
                 forward_batch_size=128):
        self.model = model
        self.large_batch = large_batch
        self.forward_batch_size = forward_batch_size
        self.N = _get_dataset_length(dataset, default=1)

        super(ModelSampler, self).__init__(dataset, reweighting)

    def _get_samples_with_scores(self, batch_size):
        assert batch_size < self.large_batch

        # Sample a large number of points in random and score them
        idxs = np.random.choice(self.N, self.large_batch)
        x, y = self.dataset.train_data[idxs]
        scores = self.model.score(x, y, batch_size=self.forward_batch_size)

        return (
            idxs,
            scores,
            (x, y)
        )


class LSTMSampler(BaseSampler):
    """Use an LSTM to predict the loss based on the previous losses of each
    sample

    Arguments
    ---------

    dataset: The dataset we want to sample from
    presample: int
               Presample that many data points using uniform sampling to
               decrease the complexity
    history: int
             How many scores per data point to keep in history
    smooth: float
            Depending on whether we are using adaptive smoothing or additive we
            either add smooth*mean or simply smooth to each importance before
            sampling
    warmup: int
            Sample unformly at first that many times to gather some history
    log: bool
         Do the regression in log space
    adaptive_smooth: bool
                     Smooth adaptively based on the mean of the scores
    forget: float
            A float less than one to used to calculate the mean of the scores
    """
    def __init__(self, dataset, reweighting, presample=2048, history=10, warmup=100,
                 log=False):
        # Initialize the history for every sample
        init = -np.log(1.0 / dataset.output_size)
        if log:
            init = np.log(init)
        self.history = np.zeros((len(dataset.train_data), history, 1))
        self.history[:, 0, 0] = init
        self.cnts = np.ones(len(dataset.train_data), dtype=np.uint8)

        # Keep some member variables
        self.presample = presample
        self.log = log
        self.warmup = warmup
        self.batches_sampled = 0

        # Create our LSTM model
        x00 = Input(shape=(history, 1))
        x10 = Input(shape=(1,))
        x0 = Masking(mask_value=0.0)(x00)
        x0 = LSTM(32, return_sequences=True, unroll=True)(x0)
        x0 = LSTM(32, unroll=True)(x0)
        x1 = Embedding(dataset.output_size, 32)(x10)
        x1 = Flatten()(x1)
        x = concatenate([x0, x1])
        y = Dense(1)(x)
        self.model = Model(inputs=[x00, x10], outputs=y)
        self.model.compile(optimizer="adam", loss="mse")

        super(LSTMSampler, self).__init__(dataset, reweighting)

    def _to_ids(self, y):
        try:
            if y.shape[1] > 1:
                return np.expand_dims(y.argmax(axis=1), -1)
        except:
            return y

    def _get_samples_with_scores(self, batch_size):
        """Use the LSTM to predict the loss of each sample"""
        # Increment our batch counter
        self.batches_sampled += 1

        # if we are still warming up just sample randomly
        if self.batches_sampled <= self.warmup:
            return (
                np.random.choice(len(self.dataset.train_data), self.presample),
                np.ones(self.presample),
                None
            )

        # Presample so that we do not run the LSTM for the whole dataset
        idxs = np.random.choice(len(self.history), self.presample)
        x, y = self.dataset.train_data[idxs]

        # Predict normalize and sample
        scores = self.model.predict(
            [self.history[idxs], self._to_ids(y)],
            batch_size=1024
        ).ravel()

        # Perform the regression in logspace if needed
        if self.log:
            np.exp(scores, scores)
        else:
            np.maximum(scores, 0, scores)

        return (
            idxs,
            scores,
            (x, y)
        )

    def update(self, idxs, x):
        # Fetch the classes for the regression
        _, y = self.dataset.train_data[idxs]

        # If we are doing the regression in logspace
        if self.log:
            x = np.log(x)

        # Train the lstm so that it can predict x given the history
        self.model.train_on_batch([self.history[idxs], self._to_ids(y)], x)

        # Update the history to include x
        full = idxs[self.cnts[idxs] == self.history.shape[1]]
        self.history[full] = np.roll(self.history[full], -1, axis=1)
        self.cnts[full] -= 1
        self.history[idxs, self.cnts[idxs], :1] = x
        self.cnts[idxs] += 1


class PerClassGaussian(BaseSampler):
    """Fit a Gaussian per class to predict the losses"""
    def __init__(self, dataset, reweighting, alpha=0.9, presample=2048):
        self.alpha = alpha
        self.presample = presample
        self.mu = np.ones(dataset.output_size) * np.log(dataset.output_size)
        self.variance = np.ones(dataset.output_size)

        super(PerClassGaussian, self).__init__(dataset, reweighting)

    def _get_samples_with_scores(self, batch_size):
        # Presample so that we do not need to compute everything
        # on the whole dataset
        idxs = np.random.choice(len(self.dataset.train_data), self.presample)
        x, y = self.dataset.train_data[idxs]
        yis = y.ravel() if y.shape[1] == 1 else y.argmax(axis=1)

        # Compute the sampling probs for each of the above presampled
        # data points
        scores = self.mu + np.sqrt(np.maximum(self.variance - self.mu**2, 0))
        scores = scores[yis]

        return (
            idxs,
            scores,
            (x, y)
        )

    def update(self, idxs, x):
        # Fetch the classes in order to model per class information
        _, y = self.dataset.train_data[idxs]
        yis = y.ravel() if y.shape[1] == 1 else y.argmax(axis=1)

        # Update the mean and variance one by one
        # TODO: Improve the following implementation
        for xi, yi in zip(x, yis):
            d = (1.0 - self.alpha) * xi
            self.mu[yi] = self.alpha * self.mu[yi] + d
            self.variance[yi] = self.alpha * self.variance[yi] + d * xi


class LSTMComparisonSampler(BaseSampler):
    """Compare LSTM and Model scores on a fixed presampled subset of the
    training data"""
    def __init__(self, dataset, lstm, model, subset=1024):
        self._idxs = np.random.choice(len(dataset.train_data), subset)
        self._x, self._y= dataset.train_data[self._idxs]
        self.lstm = lstm
        self.model = model

    def _get_samples_with_scores(self, batch_size):
        s1 = self.model.model.score(self._x, self._y)
        s2 = self.lstm.model.predict(
            [self.lstm.history[self._idxs], self.lstm._to_ids(self._y)],
            batch_size=1024
        ).ravel()
        signal("is.lstm_comparison_sampler.scores").send(zip(s1, s2))

        return self.lstm._get_samples_with_scores(batch_size)

    def update(self, idxs, x):
        return self.lstm.update(idxs, x)


class SamplerDecorator(BaseSampler):
    """Just decorate another sampler.

    Arguments
    ---------
    sampler: BaseSampler
             The sampler being decorated
    """
    def __init__(self, sampler):
        self.sampler = sampler

        super(SamplerDecorator, self).__init__(
            sampler.dataset,
            sampler.reweighting
        )

    def _get_samples_with_scores(self, batch_size):
        raise NotImplementedError()

    def update(self, idxs, results):
        self.sampler.update(idxs, results)


class AdditiveSmoothingSampler(SamplerDecorator):
    """Add a constant to all the importance scores in order to smooth them
    towards uniform

    Arguments
    ---------
    sampler: BaseSampler
             The sampler being decorated
    c: float
       A constant to add to every importance score
    """
    def __init__(self, sampler, c=1.0):
        self.c = c
        super(AdditiveSmoothingSampler, self).__init__(sampler)

    def _get_samples_with_scores(self, batch_size):
        idxs, scores, xy = self.sampler._get_samples_with_scores(batch_size)

        return (
            idxs,
            scores + self.c,
            xy
        )


class AdaptiveAdditiveSmoothingSampler(SamplerDecorator):
    """Add a percentage of the moving average of the predicted importance
    scores to smooth them towards uniform.

    Arguments
    ---------
    sampler: BaseSampler
             The sampler being decorated
    percentage: float
                Multiplied by the moving average of the importance scores to
                add to each score to smooth it
    forget: float
            Used to compute the exponential moving average mu = forget * mu +
            (1-forget) * mu_new
    """
    def __init__(self, sampler, percentage=0.5, forget=0.9):
        self.percentage = percentage
        self.forget = forget
        self.mu = 1.0  # it could be 0 it doesn't really matter
        super(AdaptiveAdditiveSmoothingSampler, self).__init__(sampler)

    def _get_samples_with_scores(self, batch_size):
        idxs, scores, xy = self.sampler._get_samples_with_scores(batch_size)

        self.mu = self.forget * self.mu + (1 - self.forget) * scores.mean()

        return (
            idxs,
            scores + self.percentage * self.mu,
            xy
        )


class PowerSmoothingSampler(SamplerDecorator):
    """Raise the importance scores to a power (less than 1) to smooth them
    towards uniform.
    
    Arguments
    ---------
    sampler: BaseSampler
             The sampler being decorated
    power: float
           The power to raise the scores to
    """
    def __init__(self, sampler, power=0.5):
        assert 0 <= power <= 1

        self.power = power
        super(PowerSmoothingSampler, self).__init__(sampler)

    def _get_samples_with_scores(self, batch_size):
        idxs, scores, xy = self.sampler._get_samples_with_scores(batch_size)

        return (
            idxs,
            scores**self.power,
            xy
        )


class OnlineBatchSelectionSampler(ModelSampler):
    """OnlineBatchSelection is the online batch creation method by Loschchilov
    & Hutter.

    See 'Online Batch Selection for Faster Training of Neural Networks'.

    Arguments
    ---------
    dataset: The dataset to sample from
    reweighting: The reweighting scheme
    model: The model to be used for scoring
    steps_per_epoch: int
                     How many batches to create before considering that an
                     epoch has passed
    recompute: int
               Recompute the scores after r minibatches seen
    s_e: tuple
         Used to compute the sampling probabilities from the ranking
    n_epochs: int
              The number of epochs, used to compute the sampling probabilities
    """
    def __init__(self, dataset, reweighting, model, large_batch=1024,
                 forward_batch_size=128, steps_per_epoch=300, recompute=2,
                 s_e=(1, 1), n_epochs=1):
        super(OnlineBatchSelectionSampler, self).__init__(
            dataset,
            reweighting,
            model,
            large_batch=large_batch,
            forward_batch_size=forward_batch_size
        )

        # The configuration of OnlineBatchSelection
        self.steps_per_epoch = steps_per_epoch
        self.recompute = recompute
        self.s_e = s_e
        self.n_epochs = n_epochs

        # Mutable variables to be updated
        self._batch = 0
        self._epoch = 0
        self._raw_scores = np.ones((len(dataset.train_data),))
        self._scores = np.ones_like(self._raw_scores)
        self._ranks = np.arange(len(dataset.train_data))

    def _get_samples_with_scores(self, batch_size):
        return (
            np.arange(len(self._ranks)),
            self._scores,
            None
        )

    def update(self, idxs, results):
        # Compute the current epoch and the current batch
        self._batch += 1
        self._epoch = 1 + self._batch / self.steps_per_epoch

        # Add the new scores to the raw_scores
        self._raw_scores[idxs] = results.ravel()

        # if it is a new epoch
        if self._batch % self.steps_per_epoch == 0:
            # For the very first batch or every 'recompute' epochs compute the
            # loss across the whole dataset
            if self.recompute > 0 and self._epoch % self.recompute == 0:
                # Extra internal batch size so that we do not load too much
                # data into memory
                scores = []
                for i in range(0, len(self.dataset.train_data), 1024*64):
                    x, y = self.dataset.train_data[i:i+1024*64]
                    scores.append(self.model.score(
                        x, y, batch_size=self.forward_batch_size
                    ))
                self._raw_scores[:] = np.hstack(scores)

            # Sort and recompute the ranks
            N = len(self.dataset.train_data)
            self._ranks[self._raw_scores.argsort()] = np.arange(N)[::-1]

            # Recompute the sample scores from the ranks
            s_e0, s_eend = self.s_e
            n_epochs = self.n_epochs
            s = s_e0 * np.exp(np.log(s_eend/s_e0)/n_epochs) ** self._epoch
            s = 1.0 / np.exp(np.log(s)/N)
            self._scores = s**self._ranks



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from keras import backend as K
from keras.layers import    \
    Activation,             \
    BatchNormalization,     \
    Convolution2D,          \
    Dense,                  \
    Dropout,                \
    ELU,                    \
    Embedding,              \
    Flatten,                \
    GlobalAveragePooling2D, \
    Input,                  \
    LSTM,                   \
    MaxPooling2D,           \
    add
from keras.models import Model, Sequential
from keras.optimizers import SGD


def build_small_nn(input_shape, output_size):
    model = Sequential([
        Dense(40, activation="tanh", input_shape=input_shape),
        Dense(40, activation="tanh"),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        loss="categorical_crossentropy",
        optimizer="adam",
        metrics=["accuracy"]
    )

    return model


def build_cnn(input_shape, output_size):
    kwargs = {
        "kernel_size": 3,
        "activation": "relu",
        "padding": "same"
    }
    model = Sequential([
        # conv1_*
        Convolution2D(64, input_shape=input_shape, **kwargs),
        BatchNormalization(),
        Convolution2D(64, **kwargs),
        BatchNormalization(),
        MaxPooling2D(pool_size=(2, 2)),
        Dropout(0.25),

        # conv2_*
        Convolution2D(128, **kwargs),
        BatchNormalization(),
        Convolution2D(128, **kwargs),
        BatchNormalization(),
        MaxPooling2D(pool_size=(2, 2)),
        Dropout(0.25),

        # conv3_*
        Convolution2D(256, **kwargs),
        BatchNormalization(),
        Convolution2D(256, **kwargs),
        BatchNormalization(),
        MaxPooling2D(pool_size=(2, 2)),
        Dropout(0.25),

        # Fully connected
        Flatten(),
        Dense(1024),
        Activation("relu"),
        Dropout(0.5),
        Dense(512),
        Activation("relu"),
        Dropout(0.5),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        loss="categorical_crossentropy",
        optimizer="adam",
        metrics=["accuracy"]
    )

    return model


def build_small_cnn(input_shape, output_size):
    model = Sequential([
        # conv1_*
        Convolution2D(32, kernel_size=3, padding="same",
                      input_shape=input_shape),
        Activation("relu"),
        Convolution2D(32, kernel_size=3, padding="same"),
        Activation("relu"),
        MaxPooling2D(pool_size=(2, 2)),
        Dropout(0.25),

        # conv2_*
        Convolution2D(64, kernel_size=3, padding="same"),
        Activation("relu"),
        Convolution2D(64, kernel_size=3, padding="same"),
        Activation("relu"),
        MaxPooling2D(pool_size=(2, 2)),
        Dropout(0.25),

        # Fully connected
        Flatten(),
        Dense(512),
        Activation("relu"),
        Dropout(0.5),
        Dense(512),
        Activation("relu"),
        Dropout(0.5),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        loss="categorical_crossentropy",
        optimizer="adam",
        metrics=["accuracy"]
    )

    return model


def build_lr(input_shape, output_size):
    model = Sequential([
        Flatten(input_shape=input_shape),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        loss="categorical_crossentropy",
        optimizer="adam",
        metrics=["accuracy"]
    )

    return model


def build_all_conv_nn(input_shape, output_size):
    """Build a small variation of the best performing network from
    'Springenberg, Jost Tobias, et al. "Striving for simplicity: The all
     convolutional net." arXiv preprint arXiv:1412.6806 (2014)' which should
     achieve approximately 91% in CIFAR-10.
    """
    kwargs = {
        "activation": "relu",
        "border_mode": "same"
    }
    model = Sequential([
        # conv1
        Convolution2D(96, 3, 3, input_shape=input_shape, **kwargs),
        BatchNormalization(),
        Convolution2D(96, 3, 3, **kwargs),
        BatchNormalization(),
        Convolution2D(96, 3, 3, subsample=(2, 2), **kwargs),
        BatchNormalization(),
        Dropout(0.25),

        # conv2
        Convolution2D(192, 3, 3, **kwargs),
        BatchNormalization(),
        Convolution2D(192, 3, 3, **kwargs),
        BatchNormalization(),
        Convolution2D(192, 3, 3, subsample=(2, 2), **kwargs),
        BatchNormalization(),
        Dropout(0.25),

        # conv3
        Convolution2D(192, 1, 1, **kwargs),
        BatchNormalization(),
        Dropout(0.25),
        Convolution2D(output_size, 1, 1, **kwargs),
        GlobalAveragePooling2D(),
        Activation("softmax")
    ])

    model.compile(
        loss="categorical_crossentropy",
        optimizer=SGD(momentum=0.9),
        metrics=["accuracy"]
    )

    return model


def build_elu_cnn(input_shape, output_size):
    """Build a variation of the CNN implemented in the ELU paper.

    https://arxiv.org/abs/1511.07289
    """
    def layers(n, channels, kernel):
        return sum(
            (
                [
                    Convolution2D(
                        channels,
                        kernel_size=kernel,
                        padding="same"
                    ),
                    ELU()
                ]
                for i in range(n)
            ),
            []
        )

    model = Sequential(
        [
            Convolution2D(384, kernel_size=3, padding="same",
                          input_shape=input_shape)
        ] +
        layers(1, 384, 3) +
        [MaxPooling2D(pool_size=(2, 2))] +
        layers(1, 384, 1) + layers(1, 384, 2) + layers(2, 640, 2) +
        [MaxPooling2D(pool_size=(2, 2))] +
        layers(1, 640, 1) + layers(3, 768, 2) +
        [MaxPooling2D(pool_size=(2, 2))] +
        layers(1, 768, 1) + layers(2, 896, 2) +
        [MaxPooling2D(pool_size=(2, 2))] +
        layers(1, 896, 3) + layers(2, 1024, 2) +
        [
            MaxPooling2D(pool_size=(2, 2)),
            Convolution2D(output_size, kernel_size=1, padding="same"),
            GlobalAveragePooling2D(),
            Activation("softmax")
        ]
    )

    model.compile(
        optimizer=SGD(momentum=0.9),
        loss="categorical_crossentropy",
        metrics=["accuracy"]
    )

    return model


def build_lstm_lm(input_shape, output_size):
    # LM datasets will report the vocab_size as output_size
    vocab_size = output_size

    model = Sequential([
        Embedding(vocab_size + 1, 64, mask_zero=True,
                  input_length=input_shape[0]),
        LSTM(256, unroll=True, return_sequences=True),
        Dropout(0.5),
        LSTM(256, unroll=True),
        Dropout(0.5),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        optimizer="adam",
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"]
    )

    return model


def build_lstm_lm2(input_shape, output_size):
    # LM datasets will report the vocab_size as output_size
    vocab_size = output_size

    model = Sequential([
        Embedding(vocab_size + 1, 128, mask_zero=True,
                  input_length=input_shape[0]),
        LSTM(384, unroll=True, return_sequences=True),
        Dropout(0.5),
        LSTM(384, unroll=True),
        Dropout(0.5),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        optimizer="adam",
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"]
    )

    return model


def build_lstm_lm3(input_shape, output_size):
    # LM datasets will report the vocab_size as output_size
    vocab_size = output_size

    model = Sequential([
        Embedding(vocab_size + 1, 128, mask_zero=True,
                  input_length=input_shape[0]),
        LSTM(650, unroll=True, return_sequences=True),
        Dropout(0.5),
        LSTM(650, unroll=True),
        Dropout(0.5),
        Dense(output_size),
        Activation("softmax")
    ])

    model.compile(
        optimizer="adam",
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"]
    )

    return model


def build_small_cnn_squared(input_shape, output_size):
    def squared_categorical_crossent(y_true, y_pred):
        return K.square(K.categorical_crossentropy(y_pred, y_true))

    model = build_small_cnn(input_shape, output_size)
    model.compile(
        optimizer=model.optimizer,
        loss=squared_categorical_crossent,
        metrics=model.metrics
    )

    return model


def wide_resnet(L, k, drop_rate=0.0):
    """Implement the WRN-L-k from 'Wide Residual Networks' BMVC 2016"""
    def wide_resnet_impl(input_shape, output_size):
        def conv(channels, strides,
                 params=dict(padding="same", use_bias=False)):
            def inner(x):
                x = BatchNormalization()(x)
                x = Activation("relu")(x)
                x = Convolution2D(channels, 3, strides=strides, **params)(x)
                x = Dropout(drop_rate)(x) if drop_rate > 0 else x
                x = BatchNormalization()(x)
                x = Activation("relu")(x)
                x = Convolution2D(channels, 3, **params)(x)
                return x
            return inner

        def resize(x, shape):
            if K.int_shape(x) == shape:
                return x
            channels = shape[3 if K.image_data_format() == "channels_last" else 1]
            strides = K.int_shape(x)[2] / shape[2]
            return Convolution2D(
                channels, 1, padding="same", use_bias=False, strides=strides
            )(x)

        def block(channels, k, n, strides):
            def inner(x):
                for i in range(n):
                    x2 = conv(channels*k, strides if i==0 else 1)(x)
                    x = add([resize(x, K.int_shape(x2)), x2])
                return x
            return inner

        # According to the paper L = 6*n+4
        n = (L-4)/6

        group0 = Convolution2D(16, 3, padding="same", use_bias=False)
        group1 = block(16, k, n, 1)
        group2 = block(32, k, n, 2)
        group3 = block(64, k, n, 2)

        x_in = x = Input(shape=input_shape)
        x = group0(x)
        x = group1(x)
        x = group2(x)
        x = group3(x)

        x = BatchNormalization()(x)
        x = Activation("relu")(x)
        x = GlobalAveragePooling2D()(x)
        x = Dense(output_size)(x)
        y = Activation("softmax")(x)

        model = Model(inputs=x_in, outputs=y)
        model.compile(
            loss="categorical_crossentropy",
            optimizer="adam",
            metrics=["accuracy"]
        )

        return model
    return wide_resnet_impl


def get(name):
    models = {
        "small_nn": build_small_nn,
        "small_cnn": build_small_cnn,
        "small_cnn_sq": build_small_cnn_squared,
        "cnn": build_cnn,
        "all_conv": build_all_conv_nn,
        "elu_cnn": build_elu_cnn,
        "lstm_lm": build_lstm_lm,
        "lstm_lm2": build_lstm_lm2,
        "lstm_lm3": build_lstm_lm3,
        "wide_resnet_16_4": wide_resnet(16, 4),
        "wide_resnet_16_4_dropout": wide_resnet(16, 4, 0.3),
        "wide_resnet_28_10": wide_resnet(28, 10),
        "wide_resnet_28_10_dropout": wide_resnet(28, 10, 0.3)
    }
    return models[name]



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""This module should provide some useful functions that I wish existed in the
functools standard package. But then again standard library developers know
better"""


placeholder = object()
x = placeholder
_ = placeholder
___ = placeholder


def partial(func, *args, **kwargs):
    """Partially define the argument list of func

    Example:
    >>> import functional as f
    >>> from operator import add
    >>> add2 = f.partial(add, f._, 2)
    >>> add2(5)
    7
    >>> add2(8)
    10
    >>> join = f.compose(' '.join, lambda *args: list(args))
    >>> create_sentence = f.partial(join, "The", f._, f._, "fox")
    >>> create_sentence("quick", "brown")
    'The quick brown fox'
    >>> import tempfile
    >>> fd, file_name = tempfile.mkstemp()
    >>> opener = f.partial(open, f._, "wb")
    >>> with opener(file_name) as out:
    ...     out.write("Hello World!")
    ...
    >>> with opener(file_name) as out:
    ...     out.write("Hello World!")
    ...
    >>> import os
    >>> os.remove(file_name)
    >>>
    """
    def inner(*args2, **kwargs2):
        # compute the kwargs to pass to the function
        fkwargs = kwargs.copy()
        fkwargs.update(kwargs2)

        # walk through the fargs list replacing each placeholder with an arg
        args2 = list(args2)
        fargs = list(args)
        for i in range(len(fargs)):
            if fargs[i] is placeholder:
                fargs[i] = args2.pop(0)
        fargs += args2

        # apply fargs and fkwargs to func
        return func(*fargs, **fkwargs)
    return inner


def compose(*functions):
    """Compose the functions with the following rules

    If a function returns a tuple the parameters are passed as positional
    arguments.

    Example:
    >>> import functional as f
    >>> add2 = f.compose(lambda x: x+1, lambda x: x+1)
    >>> add2(2)
    4
    >>> testf = f.compose(lambda x: x**2, lambda x: 2*x)
    >>> testf(1)
    4
    >>> testf = f.compose(lambda x: x**2, lambda a,b: a+b)
    >>> testf(1, 1)
    4
    >>>
    """
    def inner(*args, **kwargs):
        for f in reversed(functions):
            args = f(*args, **kwargs)
            if isinstance(args, dict):
                kwargs = args
                args = tuple()
            elif not isinstance(args, tuple):
                args = (args, )
            kwargs = {}
        if not isinstance(args, tuple) or len(args) > 1:
            return args
        else:
            return args[0]
    return inner


def call(f, *args, **kwargs):
    """Call f with args and kwargs"""
    return f(*args, **kwargs)


def attr(a):
    """Return a function that returns the attribute 'a'.

    It would be equal to partial(getattr(___, a)) but I define it separately
    for better documentation.
    """
    def inner(o):
        return getattr(o, a)
    return inner


def identity(x):
    return x



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from keras import backend as K
from keras import objectives
from keras.layers import Layer


class LossLayer(Layer):
    """LossLayer outputs the loss per sample
    
    # Arguments
        loss: The loss function to use to combine the model output and the
              target
    """
    def __init__(self, loss, **kwargs):
        self.loss = objectives.get(loss)

        super(LossLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        pass # Nothing to do

        super(LossLayer, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        # We need two inputs X and y
        assert len(input_shape) == 2

        # (None, 1) because all losses should be scalar
        return (input_shape[0][0], 1)

    def call(self, x, mask=None):
        # We need two inputs X and y
        assert len(x) == 2

        return K.expand_dims(self.loss(*x))


class GradientNormLayer(Layer):
    """GradientNormLayer aims to output the gradient norm given a list of
    parameters (whose gradient to compute) and a loss function to combine the
    two inputs.
    
    # Arguments
        parameter_list: A list of Keras variables to compute the gradient
                        norm for
        loss: The loss function to use to combine the model output and the
              target into a scalar and then compute the gradient norm
        fast: If set to True it means we know that the gradient with respect to
              each sample only affects one part of the parameter list so we can
              use the batch mode to compute the gradient
    """
    def __init__(self, parameter_list, loss, fast=False, **kwargs):
        self.parameter_list = parameter_list
        self.loss = objectives.get(loss)
        self.fast = fast

        super(GradientNormLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        pass # Nothing to do

        super(GradientNormLayer, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        # We get two inputs
        assert len(input_shape) == 2

        return (input_shape[0][0], 1)

    def call(self, x, mask=None):
        # x should be an output and a target
        assert len(x) == 2

        losses = self.loss(*x)
        if self.fast:
            grads = K.sqrt(sum([
                K.sum(K.square(g), axis=1)
                for g in K.gradients(losses, self.parameter_list)
            ]))
        else:
            nb_samples = K.shape(losses)[0]
            grads = K.map_fn(
                lambda i: self._grad_norm(losses[i]),
                K.arange(0, nb_samples),
                dtype=K.floatx()
            )

        return K.reshape(grads, (-1, 1))

    def _grad_norm(self, loss):
        grads = K.gradients(loss, self.parameter_list)
        return K.sqrt(
            sum([
                K.sum(K.square(g))
                for g in grads
            ])
        )



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""This module aims to configure the Keras tensorflow session based on
environment variables or other sources"""

from multiprocessing import cpu_count
import os

from keras import backend as K


def with_tensorflow(f):
    def inner(*args, **kwargs):
        if K.backend() == "tensorflow":
            return f(K.tf, *args, **kwargs)
    return inner

@with_tensorflow
def set_random_seed(tf, seed):
    return tf.set_random_seed(seed)


if K.backend() == "tensorflow":
    tf = K.tf

    TF_THREADS = int(os.environ.get("TF_THREADS", cpu_count()))

    config = tf.ConfigProto(
        intra_op_parallelism_threads=TF_THREADS,
        inter_op_parallelism_threads=TF_THREADS,
        device_count={"CPU": TF_THREADS}
    )
    session = tf.Session(config=config)
    K.set_session(session)



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from keras import backend as K
from keras.layers import Layer
import numpy as np


class ReweightingPolicy(object):
    """ReweightingPolicy defines how we weigh the samples given their
    importance.
    
    Each policy should provide
        1. A layer implementation for use with Keras models
        2. A python implementation for use with the samplers
    """
    def weight_layer(self):
        """Return a layer that accepts the scores and the return value of the
        sample_weights() method and produces new sample weights."""
        raise NotImplementedError()

    def sample_weights(self, idxs, scores):
        """Given the scores and the chosen indices return whatever is needed by
        the weight_layer to produce the sample weights."""
        raise NotImplementedError()

    @property
    def weight_size(self):
        """Return how many numbers per sample make up the sample weights"""
        raise NotImplementedError()


class AdjustedBiasedReweightingPolicy(ReweightingPolicy):
    """AdjustedBiasedReweightingPolicy adjusts the biased sample weights with
    the importance that has just been computed in the forward-backward pass.

    See AdjustedBiasedReweighting for details.
    """
    def __init__(self, k=1.0):
        self.k = k

    def weight_layer(self):
        return AdjustedBiasedReweighting(self.k)

    def sample_weights(self, idxs, scores):
        N = len(scores)
        S1 = scores[np.setdiff1d(np.arange(N), idxs)].sum()

        return np.tile([float(N), float(S1)], (len(idxs), 1))

    @property
    def weight_size(self):
        return 2


class BiasedReweightingPolicy(ReweightingPolicy):
    """BiasedReweightingPolicy computes the sample weights before  the
    forward-backward pass based on the sampling probabilities. It can introduce
    a bias that focuses on the hard examples when combined with the loss as an
    importance metric."""
    def __init__(self, k=1.0):
        self.k = k

    def weight_layer(self):
        return ExternalReweighting()

    def sample_weights(self, idxs, scores):
        N = len(scores)
        s = scores[idxs]
        w = scores.sum() / N / s
        w_hat = w**self.k
        w_hat *= w.dot(s) / w_hat.dot(s)

        return w_hat[:, np.newaxis]

    @property
    def weight_size(self):
        return 1


class NoReweightingPolicy(ReweightingPolicy):
    """Set all sample weights to 1."""
    def weight_layer(self):
        return ExternalReweighting()

    def sample_weights(self, idxs, scores):
        return np.ones((len(idxs), 1))

    @property
    def weight_size(self):
        return 1


class CorrectingReweightingPolicy(ReweightingPolicy):
    """CorrectingReweightingPolicy aims to scale the sample weights according
    to the mistakes of the importance predictor

    Arguments
    ---------
    k: float
       The bias power used in all other reweighting schemes
    """
    def __init__(self, k=1.0):
        self.k = k
        self._biased_reweighting = BiasedReweightingPolicy(k)

    def weight_layer(self):
        return CorrectingReweighting()

    def sample_weights(self, idxs, scores):
        w = self._biased_reweighting.sample_weights(idxs, scores)

        return np.hstack([w, scores[idxs][:, np.newaxis]])

    @property
    def weight_size(self):
        return 2


class AdjustedBiasedReweighting(Layer):
    """Implement a Keras layer that using the sum of the weights and the number
    of samples it recomputes the weights.
    
    The specifics are the following:

    Given B = {i_1, i_2, ..., i_|B|} the mini-batch idexes, \\hat{s_i} the
    predicted score of sample i and k the max-loss bias constant.

        S_1 = \\sum_{i \\notin B} \\hat{s_i}
        S_2 = \\sum_{i \\in B} s_i
        a_i^{(1)} = \\frac{1}{N} (\\frac{S1 + S2}{s_i})^k \\forall i \\in B
        a_i^{(2)} = \\frac{1}{N} \\frac{S1 + S2}{s_i} \\forall i \\in B
        t = \\frac{\\sum_{i} a_i^{(2)} s_i }{\\sum_{i} a_i^{(1)} s_i}
        a_i = t a_i^{(1)} \\forall i \\in B
    """
    def __init__(self, k=1.0, **kwargs):
        self.k = k

        super(AdjustedBiasedReweighting, self).__init__(**kwargs)

    def build(self, input_shape):
        assert isinstance(input_shape, list)
        assert len(input_shape) == 2
        assert input_shape[0][1] == 1
        assert input_shape[1][1] == 2

        super(AdjustedBiasedReweighting, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def call(self, x):
        s, s_hat = x

        # Compute the variables defined in the class comment
        S2 = K.sum(s)
        S1 = s_hat[0, 1]
        N = s_hat[0, 0]

        # Compute the unbiased weights
        a2 = (S1 + S2) / N / s

        # Compute the biased weights and the scaling factor t
        a1 = K.pow(a2, self.k)
        sT = K.transpose(s)
        t = K.dot(sT, a2) / K.dot(sT, a1)

        return K.stop_gradient([a1 * t])[0]


class ExternalReweighting(Layer):
    """Use the provided input as sample weights"""
    def build(self, input_shape):
        super(ExternalReweighting, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def call(self, x):
        return K.stop_gradient(x[1])


class CorrectingReweighting(Layer):
    """Use the provided weights and the score to correct sample weights that
    were computed with a very wrong predicted score"""
    def __init__(self, min_decrease=0, max_increase=2, **kwargs):
        self.min_decrease = min_decrease
        self.max_increase = max_increase

        super(CorrectingReweighting, self).__init__(**kwargs)

    def build(self, input_shape):
        super(CorrectingReweighting, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def call(self, x):
        s, x1 = x
        a = x1[:, :1]
        s_hat = x1[:, 1:2]

        # Rescale the weights, making sure we mostly scale down
        a_hat = a * K.clip(s_hat / s, self.min_decrease, self.max_increase)

        # Scale again so that the reported loss is comparable to the other ones
        t = 1
        #sT = K.transpose(s)
        #t = K.dot(sT, a) / K.dot(sT, a_hat)

        return K.stop_gradient([a_hat * t])[0]


UNWEIGHTED = NoReweightingPolicy()
UNBIASED = AdjustedBiasedReweightingPolicy(k=1.0)



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from collections import OrderedDict
from functools import partial
import gzip
from itertools import islice
from os import path
import pickle
from tempfile import TemporaryFile

from keras.datasets import cifar10, cifar100, mnist
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import np_utils
from keras.utils.generic_utils import Progbar
import numpy as np

from .utils.functional import compose


class cifar_sanity_check(object):
    """Create a dataset from a subset of CIFAR10 that is designed to be ideal
    for importance sampling by having a sample repeated thousands of times"""
    def __init__(self, classes=(3, 5), replicate=30000, replicate_idx=42):
        assert len(classes) > 1
        self.classes = classes
        self.replicate = replicate
        self.replicate_idx = replicate_idx

    def load_data(self):
        # Load the original data
        (X_train, y_train), (X_test, y_test) = cifar10.load_data()

        # Get only the classes given
        idxs_train = np.arange(len(y_train))[
            np.logical_or(*[y_train == c for c in self.classes]).ravel()
        ]
        idxs_test = np.arange(len(y_test))[
            np.logical_or(*[y_test == c for c in self.classes]).ravel()
        ]
        X_train = X_train[idxs_train]
        y_train = y_train[idxs_train]
        X_test = X_test[idxs_test]
        y_test = y_test[idxs_test]
        for i, c in enumerate(self.classes):
            y_train[y_train == c] = i
            y_test[y_test == c] = i

        # Replicate on document in the training set
        x, y = X_train[self.replicate_idx], y_train[self.replicate_idx]
        x = np.tile(x, (self.replicate, 1, 1, 1))
        y = np.tile(y, (self.replicate, 1))

        return (
            (np.vstack([X_train, x]), np.vstack([y_train, y])),
            (X_test, y_test)
        )


class canevet_icml2016_nn(object):
    """Approximates the artifical dataset used in [1] section 4.2.

    [1] Can\'evet, Olivier, et al. "Importance Sampling Tree for Large-scale
        Empirical Expectation." Proceedings of the International Conference on
        Machine Learning (ICML). No. EPFL-CONF-218848. 2016.
    """
    def __init__(self, N=8192, test_split=0.33, smooth=40):
        self.N = N
        self.test_split = test_split
        self.smooth = smooth

    def load_data(self):
        # Create the data using magic numbers to approximate the figure in
        # canevet_icml2016
        x = np.linspace(0, 1, self.N).astype(np.float32)
        ones = np.ones_like(x).astype(int)
        boundary = np.sin(4*(x + 0.5)**5)/3 + 0.5

        data = np.empty(shape=[self.N, self.N, 3], dtype=np.float32)
        data[:, :, 0] = 1-x
        for i in range(self.N):
            data[i, :, 1] = 1-x[i]
            data[i, :, 2] = 1 / (1 + np.exp(self.smooth*(x - boundary[i])))
            data[i, :, 2] = np.random.binomial(ones, data[i, :, 2])
        data = data.reshape(-1, 3)
        np.random.shuffle(data)

        # Create train and test arrays
        split = int(len(data)*self.test_split)
        X_train = data[:-split, :2]
        y_train = data[:-split, 2]
        X_test = data[-split:, :2]
        y_test = data[-split:, 2]

        return (X_train, y_train), (X_test, y_test)


class BaseDataset(object):
    class _DataProxy(object):
        def __init__(self, dataset, subset):
            self.data = getattr(dataset, "_%s_data" % (subset,))
            self.size = getattr(dataset, "_%s_size" % (subset,))

        def __getitem__(self, idxs):
            return self.data(idxs)

        def __len__(self):
            return self.size()

    def _train_data(self, idxs=slice(None)):
        """Return the training data in the form (x, y)"""
        raise NotImplementedError()

    def _train_size(self):
        """Training data length"""
        return len(self._train_data()[0])

    def _test_data(self, idxs=slice(None)):
        """Return the testing data in the form (x, y)"""
        raise NotImplementedError()

    def _test_size(self):
        """Test data length"""
        return len(self._test_data()[0])

    @property
    def train_data(self):
        return self._DataProxy(self, "train")

    @property
    def test_data(self):
        return self._DataProxy(self, "test")

    @property
    def shape(self):
        """Return the shape of the samples"""
        raise NotImplementedError()

    @property
    def output_size(self):
        """Return the number of outputs"""
        raise NotImplementedError()



class InMemoryDataset(BaseDataset):
    """A dataset that fits in memory and is simply 4 numpy arrays (x, y) *
    (train, test)"""
    def __init__(self, X_train, y_train, X_test, y_test, categorical=True):
        self._x_train = X_train
        self._x_test = X_test

        # are the targets to be made one hot vectors
        if categorical:
            self._y_train = np_utils.to_categorical(y_train)
            self._y_test = np_utils.to_categorical(y_test)
            self._output_size = self._y_train.shape[1]

        # handle sparse output classification
        elif issubclass(y_train.dtype.type, np.integer):
            self._y_train = y_train
            self._y_test = y_test
            self._output_size = self._y_train.max() + 1  # assume 0 based indexes

        # not classification, just copy them
        else:
            self._y_train = y_train
            self._y_test = y_test
            self._output_size = self._y_train.shape[1]

    def _train_data(self, idxs=slice(None)):
        return self._x_train[idxs], self._y_train[idxs]

    def _test_data(self, idxs=slice(None)):
        return self._x_test[idxs], self._y_test[idxs]

    @property
    def shape(self):
        return self._x_train.shape[1:]

    @property
    def output_size(self):
        return self._output_size

    @classmethod
    def from_loadable(cls, dataset):
        (a, b), (c, d) = dataset.load_data()
        return cls(a, b, c, d)


class InMemoryImageDataset(InMemoryDataset):
    """Make sure that the in memory dataset has 4 dimensions and is normalized
    to [0, 1]"""
    def __init__(self, X_train, y_train, X_test, y_test):
        # Expand the dims and make sure the shapes are correct image shapes
        if len(X_train.shape) < 4:
            X_train = np.expand_dims(X_train, axis=-1)
            X_test = np.expand_dims(X_test, axis=-1)
        assert X_train.shape[1:] == X_test.shape[1:]
        assert len(X_train.shape) == 4

        # Normalize to [0, 1]
        X_train = X_train.astype(np.float32) / X_train.max()
        X_test = X_test.astype(np.float32) / X_test.max()

        super(InMemoryImageDataset, self).__init__(
            X_train,
            y_train,
            X_test,
            y_test
        )


CIFAR10 = partial(InMemoryImageDataset.from_loadable, cifar10)
CIFAR100 = partial(InMemoryImageDataset.from_loadable, cifar100)
MNIST = partial(InMemoryImageDataset.from_loadable, mnist)
CIFARSanityCheck = compose(
    InMemoryImageDataset.from_loadable,
    cifar_sanity_check
)
CanevetICML2016 = compose(InMemoryDataset.from_loadable, canevet_icml2016_nn)


class GeneratorDataset(BaseDataset):
    """GeneratorDataset wraps a generator (or two) and partially implements the
    BaseDataset interface."""
    def __init__(self, train_data, test_data=None, test_data_length=None):
        self._train_data_gen = train_data
        self._test_data_gen = test_data
        self._test_data_len = test_data_length

        # Determine the shapes and sizes
        x, y = next(self._train_data_gen)
        self._shape = x.shape[1:]
        self._output_size = y.shape[1] if len(y.shape) > 1 else 1

    def _get_count(self, idxs):
        if isinstance(idxs, slice):
            # Use 2**32 as infinity
            start, stop, step = idxs.indices(2**32)
            return (stop - start)/step
        elif isinstance(idxs, (list, np.ndarray)):
            return len(idxs)
        elif isinstance(idxs, int):
            return 1
        else:
            raise IndexError("Invalid indices passed to dataset")

    def _get_n_batches(self, generator, n_batches):
        batches = list(islice(generator, n_batches))

        return tuple(map(np.vstack, zip(*batches)))

    def _get_at_least_n(self, generator, n):
        cnt = 0
        batches = []
        while cnt < n:
            batch = next(generator)
            cnt += len(batch[1])
            batches.append(batch)

        return tuple(map(np.vstack, zip(*batches)))

    def _train_data(self, idxs=slice(None)):
        N = self._get_count(idxs)
        x, y = self._get_at_least_n(self._train_data_gen, N)

        return x[:N], y[:N]

    def _train_size(self):
        raise RuntimeError("This dataset has no size")

    def _test_data(self, idxs=slice(None)):
        # No test data
        if self._test_data_gen is None:
            raise RuntimeError("This dataset has no test data")

        # Test data are all in memory
        if isinstance(self._test_data_gen, (tuple, list, np.ndarray)):
            x, y = self._test_data_gen
            return x[idxs], y[idxs]

        # Test data are provided via a generator
        N = min(self._test_data_len, self._get_count(idxs))
        x, y = self._get_at_least_n(self._test_data_gen, N)

        return x[:N], y[:N]

    def _test_size(self):
        # No test data
        if self._test_data_gen is None:
            raise RuntimeError("This dataset has no test data")

        # Test data are all in memory
        if isinstance(self._test_data_gen, (tuple, list, np.ndarray)):
            x, y = self._test_data_gen
            return len(x)

        # Test data are provided via a generator
        return self._test_data_len

    @property
    def shape(self):
        return self._shape

    @property
    def output_size(self):
        return self._output_size


class AugmentedImages(BaseDataset):
    def __init__(self, dataset, augmentation_params, N=None):
        # Initialize member variables
        self.dataset = dataset
        self.augmentation_params = augmentation_params
        self.N = len(self.dataset.train_data) * 10 if N is None else N
        assert len(self.dataset.shape) == 3

        # Allocate space for the augmented data
        self._x = np.memmap(
            TemporaryFile(),
            dtype=np.float32,
            shape=(self.N,) + self.dataset.shape
        )
        self._y = np.zeros((self.N, self.dataset.output_size))

        # Train a generator and generate all the data
        generator = ImageDataGenerator(**self.augmentation_params)
        x, y = self.dataset.train_data[:]
        generator.fit(x)
        start = 0
        for bx, by in generator.flow(x, y, batch_size=128):
            end = min(self.N, start+len(bx))
            self._x[start:end] = bx[:end-start]
            self._y[start:end] = by[:end-start]

            start = end
            if start >= self.N:
                break

    def _train_data(self, idxs=slice(None)):
        return self._x[idxs], self._y[idxs]

    def _test_data(self, idxs=slice(None)):
        return self.dataset.test_data[idxs]

    @property
    def shape(self):
        return self.dataset.shape

    @property
    def output_size(self):
        return self.dataset.output_size


class OntheflyAugmentedImages(BaseDataset):
    """Use a Keras ImageDataGenerator to augment images on the fly in a
    determenistic way."""
    def __init__(self, dataset, augmentation_params, N=None, random_state=0,
                 cache_size=None):
        # Initialize some member variables
        self.dataset = dataset
        self.generator = ImageDataGenerator(**augmentation_params)
        self.N = N or (len(self.dataset.train_data) * 10)
        self.random_state = random_state
        assert len(self.dataset.shape) == 3

        # Figure out the base images for each of the augmented ones
        self.idxs = np.random.choice(
            len(self.dataset.train_data),
            self.N
        )

        # Fit the generator
        self.generator.fit(self.dataset.train_data[:][0])

        # Standardize the test data
        self._x_test = np.copy(self.dataset.test_data[:][0])
        self._x_test = self.generator.standardize(self._x_test)
        self._y_test = self.dataset.test_data[:][1]

        # Create an LRU cache to speed things up a bit for the transforms
        cache_size = cache_size or len(self.dataset.train_data)
        self.cache = OrderedDict([(-i,i) for i in range(cache_size)])
        self.cache_data = np.empty(
            shape=(cache_size,) + self.dataset.shape,
            dtype=np.float32
        )

    def _transform(self, idx, x):
        # if it is not cached add it
        if idx not in self.cache:
            # Remove the first in and add the new idx (i is the offset in
            # cache_data)
            _, i = self.cache.popitem(last=False)
            self.cache[idx] = i

            # Do the transformation and add it to the data
            np.random.seed(idx + self.random_state)
            x = self.generator.random_transform(x)
            x = self.generator.standardize(x)
            self.cache_data[i] = x

        # and if it is update it as the most recently used
        else:
            self.cache[idx] = self.cache.pop(idx)

        return self.cache_data[self.cache[idx]]

    def _train_data(self, idxs=slice(None)):
        # Make sure we accept everything that numpy accepts as indices
        idxs = np.arange(self.N)[idxs]

        # Get the original images and then transform them
        x, y = self.dataset.train_data[self.idxs[idxs]]
        x_hat = np.copy(x)
        random_state = np.random.get_state()
        for i, idx in enumerate(idxs):
            x_hat[i] = self._transform(idx, x_hat[i])
        np.random.set_state(random_state)

        return x_hat, y

    def _test_data(self, idxs=slice(None)):
        return self._x_test[idxs], self._y_test[idxs]

    def _train_size(self):
        return self.N

    @property
    def shape(self):
        return self.dataset.shape

    @property
    def output_size(self):
        return self.dataset.output_size


class PennTreeBank(BaseDataset):
    """Load the PennTreebank from Tomas Mikolov's format expected in the
    default Keras directory."""
    def __init__(self, context, ptb_path=None, val=True, verbose=True,
                 cache=True):
        if ptb_path is None:
            ptb_path = path.expanduser("~/.keras/datasets/ptb")

        if val:
            test = "valid"
        else:
            test = "test"

        # Cache the dataset for faster subsequent loads
        cache_path = path.join(ptb_path, "ptb.train-%s.pickle.gz" % (test,))

        if not path.exists(cache_path):
            with open(path.join(ptb_path, "ptb.train.txt")) as f:
                train = [l.split() + ['<EOS>'] for l in f]
            with open(path.join(ptb_path, "ptb.%s.txt" % (test,))) as f:
                test = [l.split() + ['<EOS>'] for l in f]
            V = np.array(sorted({w for l in train for w in l}))
            N = max(max(map(len, train)), max(map(len, test)))

            # No need to have context bigger than the biggest sentence
            context = min(context, N)

            # Allocate memory
            x_train = np.empty((0, context), dtype=np.int32)
            y_train = np.empty((0, 1), dtype=np.int32)
            x_test = np.empty_like(x_train)
            y_test = np.empty_like(y_train)

            # Encode the strings to numbers
            if verbose:
                prog = Progbar(len(train) + len(test))
            for i, s in enumerate(train):
                xi, yi = self._encode(s, V, context)
                x_train = np.vstack([x_train, xi])
                y_train = np.vstack([y_train, yi])
                if verbose and i % 100 == 0:
                    prog.update(i)
            for i, s in enumerate(test):
                xi, yi = self._encode(s, V, context)
                x_test = np.vstack([x_test, xi])
                y_test = np.vstack([y_test, yi])
                if verbose and i % 100 == 0:
                    prog.update(len(train) + i)
            if verbose:
                prog.update(len(train) + len(test))

            with gzip.open(cache_path, "wb") as f:
                pickle.dump(
                    {
                        "train": (x_train, y_train),
                        "test": (x_test, y_test),
                        "vocab": V
                    },
                    f,
                    protocol=2
                )

        # Read the dataset from the cached binary file
        with gzip.open(cache_path) as f:
            data = pickle.load(f)
            self._x_train, self._y_train = data["train"]
            self._x_test, self._y_test = data["test"]
            self.V = data["vocab"]

    def _encode(self, s, V, context):
        """
        Arguments
        ----------
            s: Sentence as a list of strings
            V: Vocabulary as a np array of strings
            context: The maximum length of previous words to include
        """
        idxs = np.searchsorted(V, s)
        x = np.zeros((len(s)-1, context), dtype=np.int32)
        y = np.zeros((len(s)-1, 1), np.int32)
        for i in range(1, len(s)):
            x[i-1, :i] = idxs[:i][-context:] + 1  # 0 means missing value
            y[i-1] = idxs[i]

        return x, y

    def _train_data(self, idxs=slice(None)):
        return self._x_train[idxs], self._y_train[idxs]

    def _test_data(self, idxs=slice(None)):
        return self._x_test[idxs], self._y_test[idxs]

    @property
    def shape(self):
        return self._x_train.shape[1:]

    @property
    def output_size(self):
        return len(self.V)



#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from functools import partial

from blinker import signal
from keras import backend as K
from keras.layers import Activation, Input, Lambda, multiply
from keras.metrics import categorical_accuracy, binary_accuracy, \
    get as get_metric
import numpy as np
from transparent_keras import TransparentModel

from reweighting import UNBIASED
from score_layers import GradientNormLayer, LossLayer
from utils.functional import compose


def generic_accuracy(y_true, y_pred):
    if K.int_shape(y_pred)[1] == 1:
        return binary_accuracy(y_true, y_pred)
    else:
        return categorical_accuracy(y_true, y_pred)


def MetricLayer(metric_func):
    # Special care for accuracy because keras treats it specially
    if "accuracy" in metric_func:
        metric_func = generic_accuracy
    metric_func = compose(K.expand_dims, get_metric(metric_func))

    return Lambda(lambda inputs: metric_func(*inputs), output_shape=(None, 1))


def _get_scoring_layer(score, y_true, y_pred, loss="categorical_crossentropy",
                       layer=None):
    """Get a scoring layer that computes the score for each pair of y_true,
    y_pred"""
    assert score in ["loss", "gnorm", "acc"]

    if score == "loss":
        return LossLayer(loss)([
            y_true,
            y_pred
        ])
    elif score == "gnorm":
        return GradientNormLayer(
            layer.output,
            loss,
            fast=True
        )([
            y_true,
            y_pred
        ])
    elif score == "acc":
        return LossLayer(categorical_accuracy)([
            y_true,
            y_pred
        ])


class ModelWrapper(object):
    """The goal of the ModelWrapper is to take a NN and add some extra layers
    that produce a score, a loss and the sample weights to perform importance
    sampling."""
    def evaluate(self, x, y, batch_size=128):
        bs = batch_size
        result = np.mean(np.vstack([
            self.evaluate_batch(x[s:s+bs], y[s:s+bs])
            for s in range(0, len(x), bs)
        ]), axis=0)

        signal("is.evaluation").send(result)
        return result

    def score(self, x, y, batch_size=128):
        bs = batch_size
        result = np.hstack([
            self.score_batch(x[s:s+bs], y[s:s+bs]).T
            for s in range(0, len(x), bs)
        ]).T

        signal("is.score").send(result)
        return result

    def set_lr(self, lr):
        """Set the learning rate of the wrapped models.

        We try to set the learning rate on a member variable model and a member
        variable small. If we do not find a member variable model we raise a
        NotImplementedError
        """
        try:
            K.set_value(
                self.model.optimizer.lr,
                lr
            )
        except AttributeError:
            raise NotImplementedError()

        try:
            K.set_value(
                self.small.optimizer.lr,
                lr
            )
        except AttributeError:
            pass

    def evaluate_batch(self, x, y):
        raise NotImplementedError()

    def score_batch(self, x, y):
        raise NotImplementedError()

    def train_batch(self, x, y, w):
        raise NotImplementedError()


class ModelWrapperDecorator(ModelWrapper):
    def __init__(self, model_wrapper, implemented_attributes=set()):
        self.model_wrapper = model_wrapper
        self.implemented_attributes = (
            implemented_attributes | set(["model_wrapper"])
        )

    def __getattribute__(self, name):
        _getattr = object.__getattribute__
        implemented_attributes = _getattr(self, "implemented_attributes")
        if name in implemented_attributes:
            return _getattr(self, name)
        else:
            model_wrapper = _getattr(self, "model_wrapper")
            return getattr(model_wrapper, name)


class OracleWrapper(ModelWrapper):
    AVG_LOSS = 0
    LOSS = 1
    WEIGHTED_LOSS = 2
    SCORE = 3
    METRIC0 = 4

    def __init__(self, model, reweighting, score="loss"):
        self.model = self._augment_model(model, score, reweighting)
        self.reweighting = reweighting

    def _augment_model(self, model, score, reweighting):
        # Extract some info from the model
        loss = model.loss
        optimizer = model.optimizer.__class__(**model.optimizer.get_config())
        output_shape = K.int_shape(model.output)[1:]
        if isinstance(loss, str) and loss.startswith("sparse"):
            output_shape = output_shape[:-1] + (1,)

        # Make sure that some stuff look ok
        assert not isinstance(loss, list)

        # We need to create two more inputs
        #   1. the targets
        #   2. the predicted scores
        y_true = Input(shape=output_shape)
        pred_score = Input(shape=(reweighting.weight_size,))

        # Create a loss layer and a score layer
        loss_tensor = LossLayer(loss)([y_true, model.output])
        last_layer = -2 if isinstance(model.layers[-1], Activation) else -1
        score_tensor = _get_scoring_layer(
            score,
            y_true,
            model.output,
            loss,
            model.layers[last_layer]
        )

        # Create the sample weights
        weights = reweighting.weight_layer()([score_tensor, pred_score])

        # Create the output
        weighted_loss = multiply([loss_tensor, weights])

        # Create the metric layers
        metrics = model.metrics or []
        metrics = [
            MetricLayer(metric)([y_true, model.output])
            for metric in metrics
        ]

        # Finally build, compile, return
        new_model = TransparentModel(
            inputs=[model.input, y_true, pred_score],
            outputs=[weighted_loss],
            observed_tensors=[loss_tensor, weighted_loss, score_tensor] + metrics
        )
        new_model.compile(
            optimizer=optimizer,
            loss=lambda y_true, y_pred: y_pred
        )

        return new_model

    def evaluate_batch(self, x, y):
        dummy_weights = np.ones((x.shape[0], self.reweighting.weight_size))
        dummy_target = np.zeros((x.shape[0], 1))
        outputs = self.model.test_on_batch([x, y, dummy_weights], dummy_target)

        signal("is.evaluate_batch").send(outputs)

        return np.hstack([outputs[self.LOSS]] + outputs[self.METRIC0:])

    def score_batch(self, x, y):
        dummy_weights = np.ones((x.shape[0], self.reweighting.weight_size))
        dummy_target = np.zeros((x.shape[0], 1))
        outputs = self.model.test_on_batch([x, y, dummy_weights], dummy_target)

        return outputs[self.SCORE].ravel()

    def train_batch(self, x, y, w):
        # create a dummy target to please keras
        dummy_target = np.zeros((x.shape[0], 1))

        # train on a single batch
        outputs = self.model.train_on_batch([x, y, w], dummy_target)

        # Add the outputs in a tuple to send to whoever is listening
        result = (
            outputs[self.WEIGHTED_LOSS],
            outputs[self.METRIC0:],
            outputs[self.SCORE]
        )
        signal("is.training").send(result)

        return result



#!/usr/bin/env python
#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Setup importance-sampling"""

from itertools import dropwhile
from os import path
from setuptools import find_packages, setup


def collect_docstring(lines):
    """Return document docstring if it exists"""
    lines = dropwhile(lambda x: not x.startswith('"""'), lines)
    doc = ""
    for line in lines:
        doc += line
        if doc.endswith('"""\n'):
            break

    return doc[3:-4].replace("\r", "").replace("\n", " ")


def collect_metadata():
    meta = {}
    with open(path.join("importance_sampling","__init__.py")) as f:
        lines = iter(f)
        meta["description"] = collect_docstring(lines)
        for line in lines:
            if line.startswith("__"):
                key, value = map(lambda x: x.strip(), line.split("="))
                meta[key[2:-2]] = value[1:-1]

    return meta

def setup_package():
    with open("README.rst") as f:
        long_description = f.read()
    meta = collect_metadata()
    setup(
        name="keras-importance-sampling",
        version=meta["version"],
        description=meta["description"],
        long_description=long_description,
        maintainer=meta["maintainer"],
        maintainer_email=meta["email"],
        url=meta["url"],
        license=meta["license"],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Topic :: Scientific/Engineering",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
        ],
        packages=find_packages(exclude=["docs", "tests", "scripts", "examples"]),
        install_requires=["keras>=2", "blinker", "numpy", "transparent-keras"]
    )


if __name__ == "__main__":
    setup_package()
